# Copyright (c) 2022-2023 DomainTools LLC
# Copyright (c) 2015-2016, 2019, 2021 by Farsight Security, Inc.
# This work is unpublished, and unauthorized reproduction is prohibited.
# Farsight Security, Inc. owns all rights to this work and intends that
# it be held confidential, preserved as a trade secret, and protected as
# an unpublished copyrighted work.

import base64
import hashlib
import logging
from threading import Timer, Event
from typing import BinaryIO, Optional

from .util import iterfileobj, IO_LENGTH

logger = logging.getLogger(__name__)


class DigestError(Exception):
    pass


class DigestChecker:
    def __init__(self, algorithm, init_chunk=[]):
        logger.debug('algorithm=%s', algorithm)

        self._stop_fill: bool = False
        self._fill_thread: Optional[Timer] = None
        self._ro_file: Optional[BinaryIO] = None
        self._set_digest_obj(algorithm)
        for c in init_chunk:
            self.update(c)

    def _set_digest_obj(self, algorithm):
        if algorithm is None:
            raise DigestError('No algorithm provided. Skipping digest check.')
        elif algorithm.lower() in ('sha-224', 'sha224'):
            self.digest_obj = hashlib.sha224()
        elif algorithm.lower() in ('sha-256', 'sha256'):
            self.digest_obj = hashlib.sha256()
        elif algorithm.lower() in ('sha-384', 'sha384'):
            self.digest_obj = hashlib.sha384()
        elif algorithm.lower() in ('sha-512', 'sha512'):
            self.digest_obj = hashlib.sha512()
        else:
            raise DigestError('Unsupported algorithm: {}'.format(algorithm))

    def update(self, chunk):
        self.digest_obj.update(chunk)

    def check(self, digest):
        real_digest = base64.b64encode(self.digest_obj.digest())
        digest_bytes = digest.encode('utf-8')
        if real_digest != digest_bytes:
            raise DigestError('Digest mismatch: {} != {}'.format(real_digest, digest_bytes))

    def _fill(self, part_file: BinaryIO, new_data) -> None:
        logger.debug("Filling digest checker...")
        for c in iterfileobj(part_file):
            self.update(c)
        logger.debug("Finished fill from 0 at %s bytes", part_file.tell())

        new_data.clear()
        logger.debug("Now reading new data")
        while new_data.wait(1) or not self._stop_fill:
            buf: Optional[bytes] = part_file.read(IO_LENGTH)
            if not buf:
                continue
            self.update(buf)
            new_data.clear()
        logger.debug("Finished fill at %s bytes", part_file.tell())

    def start_fill(self, file_path: str, new_data: Event) -> None:
        self._ro_file = open(file_path, "rb")
        self._fill_thread = Timer(1.0, self._fill, args=(self._ro_file, new_data,))
        self._fill_thread.start()

    def stop_fill(self) -> int:
        res: int = 0
        if not self._ro_file:
            self._fill_thread = None
            return res

        self._stop_fill = True
        if self._fill_thread:
            self._fill_thread.join()

        res = self._ro_file.tell()
        if not self._ro_file.closed:
            self._ro_file.close()

        # Tidy up in case this is called more than once
        self._fill_thread = None
        self._ro_file = None
        return res


def digest_extension(algorithm):
    if algorithm.lower() in ('sha-224', 'sha224'):
        return 'sha224'
    elif algorithm.lower() in ('sha-256', 'sha256'):
        return 'sha256'
    elif algorithm.lower() in ('sha-384', 'sha384'):
        return 'sha384'
    elif algorithm.lower() in ('sha-512', 'sha512'):
        return 'sha512'
    else:
        raise DigestError('Unknown algorithm: {}'.format(algorithm))


DIGEST_EXTENSIONS = ('sha224', 'sha256', 'sha384', 'sha512')
