# Copyright (c) 2022 DomainTools LLC
# Copyright (c) 2015-2016, 2019, 2021 by Farsight Security, Inc.
# This work is unpublished, and unauthorized reproduction is prohibited.
# Farsight Security, Inc. owns all rights to this work and intends that
# it be held confidential, preserved as a trade secret, and protected as
# an unpublished copyrighted work.

from io import BytesIO
import http.client
import os
import shutil
import tempfile
import time
import unittest
import urllib.request, urllib.parse, urllib.error
import urllib.response
import socket

from dnstable_manager import get_config, DNSTableManager
from dnstable_manager.download import DownloadManager
import jsonschema

def get_uri(obj):
    if isinstance(obj, urllib.request.Request):
        return obj.get_full_url()
    return obj

class TestGetConfig(unittest.TestCase):
    def test_get_config_default(self):
        with self.assertRaises(jsonschema.ValidationError):
            get_config()

class TestDNSTableManager(unittest.TestCase):
    @staticmethod
    def noop(self, *args, **kwargs): pass

    def setUp(self):
        self.orig_urlopen = urllib.request.urlopen
        self.orig_sleep = time.sleep
        self.td = tempfile.mkdtemp(prefix='test-dnstable-manager.')

    def tearDown(self):
        urllib.request.urlopen = self.orig_urlopen
        time.sleep = self.orig_sleep
        shutil.rmtree(self.td, ignore_errors=True)

    def test_run(self):
        uri_base = 'http://example.com'
        fileset_uri = '{}/dns.fileset'.format(uri_base)
        fileset = (
            'dns.2014.Y.mtbl',
            'dns.201501.M.mtbl',
            'dns.20150201.W.mtbl',
            'dns.20150208.D.mtbl',
            'dns.20150209.0000.H.mtbl',
            'dns.20150209.0100.X.mtbl',
            'dns.20150209.0110.m.mtbl'
            )

        def my_urlopen(obj, timeout=None):
            uri = get_uri(obj)

            msg = http.client.HTTPMessage(BytesIO())

            if uri == fileset_uri:
                return urllib.response.addinfourl(BytesIO('\n'.join(fileset + ('',)).encode('utf-8')),
                        msg, uri)
            else:
                self.assertTrue(uri.startswith('{}/'.format(uri_base)))
                return urllib.response.addinfourl(BytesIO('{}'.format(uri[1+len(uri_base):]).encode('utf-8')),
                        msg, uri)
        urllib.request.urlopen = my_urlopen

        class Success(Exception): pass
        def my_sleep(timeout):
            self.orig_sleep(0.1)
            time.sleep = my_sleep_done
        def my_sleep_done(timeout):
            self.orig_sleep(0.1)
            raise Success
        time.sleep = my_sleep

        d = DownloadManager()
        d.start()
        m = DNSTableManager(fileset_uri, self.td, download_manager=d, digest_required=False)
        self.assertRaises(Success, m.run)
        self.orig_sleep(1)
        for fn in fileset:
            with open(os.path.join(self.td, fn)) as fs:
                self.assertEqual(fs.read(), fn)
        d.stop(blocking=True)
    
    def test_run_with_socket_timeout(self):
        self.counter = 0
        uri_base = 'http://example.com'
        fileset_uri = '{}/dns.fileset'.format(uri_base)
        fileset = 'dns.2014.Y.mtbl'

        data = 'aaa\naaaa\naaaa\nbbb\nbbb\nbbb\n'.encode('utf-8')
        content_length = len(data)
        chunks = content_length // 2

        def my_urlopen(obj, timeout=None):
            uri = get_uri(obj)

            msg = http.client.HTTPMessage()

            if uri == fileset_uri:
                return urllib.response.addinfourl(BytesIO((fileset + '\n').encode('utf-8')),
                        msg, uri)
            else:
                self.counter += 1
                self.assertTrue(uri.startswith('{}/'.format(uri_base)))
                
                if self.counter == 1:
                    msg.add_header('Content-Length', '{}'.format(content_length))
                    # Send only half the data
                    send_data = data[0:chunks]
                elif self.counter  == 2:
                    # Raise error on 2nd attempt
                    raise socket.timeout('Fake socket timeout')
                else:
                    # Finally, send the other half
                    msg.add_header('Content-Length', '{}'.format(chunks))
                    send_data = data[chunks:content_length]
                return urllib.response.addinfourl(BytesIO(send_data), msg, uri)

        urllib.request.urlopen = my_urlopen

        d = DownloadManager(retry_timeout=0)
        d.start()
        m = DNSTableManager(fileset_uri, self.td, download_manager=d, digest_required=False,
            retry_timeout=0)
        m.start()
        time.sleep(1)

        path_to_check = os.path.join(self.td, fileset)
        for _ in range(3):
            if not os.path.isfile(path_to_check):
                time.sleep(1)
                continue
            m.thread.terminate()
            with open(path_to_check, 'rb') as fs:
                self.assertEqual(fs.read(), data)
                return
        self.fail('{} does not exist'.format(path_to_check))

    def test_clean_tempfiles(self):
        m = DNSTableManager(os.path.join('file://', self.td), self.td, base='dns', download_manager=None)
        closed_file = os.path.join(self.td, '.dns.2000.Y.mtbl.XXXXXX')
        opened_file = os.path.join(self.td, '.dns.2001.Y.mtbl.XXXXXX')
        open(closed_file, 'w').close()
        of = open(opened_file, 'w')

        m.clean_tempfiles()
        of.close()

        self.assertTrue(os.path.exists(opened_file))
        self.assertFalse(os.path.exists(closed_file))

    def test_clean_tempfiles_unlink_open(self):
        m = DNSTableManager(os.path.join('file://', self.td), self.td, base='dns', download_manager=None)
        opened_file = os.path.join(self.td, '.dns.2001.Y.mtbl.XXXXXX')
        of = open(opened_file, 'w')

        m.clean_tempfiles(unlink_open=True)
        of.close()

        self.assertFalse(os.path.exists(opened_file))
