# Copyright (c) 2022-2023 DomainTools LLC
# Copyright (c) 2015-2016, 2019, 2021 by Farsight Security, Inc.
# This work is unpublished, and unauthorized reproduction is prohibited.
# Farsight Security, Inc. owns all rights to this work and intends that
# it be held confidential, preserved as a trade secret, and protected as
# an unpublished copyrighted work.

import base64
from io import BytesIO
import unittest
import os
import threading
import tempfile
import time

import dnstable_manager.util as du

def _iterfileobj(data, length):
    def t(self):
        src = BytesIO(data)
        dst = BytesIO()

        for chunk in du.iterfileobj(src, length=length):
            self.assertLessEqual(len(chunk), length)
            self.assertGreater(len(chunk), 0)
            dst.write(chunk)

        self.assertEqual(dst.getvalue(), data)
    return t

class TestIterFileObj(unittest.TestCase): pass

for length in (2, 32, 1024, 16*1024):
    data = base64.b64decode('qoiMsxenAEPPiamMzviaf13rCA6s2xtJrYjGBMZ=') * (length+1)
    assert(len(data) % length > 0)
    setattr(TestIterFileObj, 'test_iterfileobj(len={})'.format(length), _iterfileobj(data, length))

class TestMergeDict(unittest.TestCase):
    def test_merge_dict_abstract(self):
        a1 = {'a': 'a1', 'b': 'b1', 'c': 'c1', 'd': None, 'e': {'aa': 'aa1'}}
        b2 = {'a': 'a2', 'c': 'c2', 'd': 'd2', 'e': {'aa': {'aaa': 'aaa2'}}}
        c3 = {'b': 'b3', 'e': None, 'f': {'aa': 'aa3', 'bb': 'bb3'}}
        expected = {
            'a': 'a2',
            'b': 'b3',
            'c': 'c2',
            'd': 'd2',
            'e': None,
            'f': {'aa': 'aa3', 'bb': 'bb3'}
        }
        actual = du.merge_dicts([a1, b2, c3])
        self.assertEqual(actual, expected)

    def test_merge_dict_realistic(self):
        default = {
            'manager': {
                'log_level': 'ERROR',
                'syslog': False,
                'syslog_facility': 'USER',
                'clean_tempfiles': {
                    'on_startup': False,
                    'on_exit': False
                }
            },
            'downloader': {
                'max_downloads': 4,
                'download_timeout': 60,
                'retry_timeout': 60,
                'ssl_ca_file': '/etc/ssl/certs/ca-certificates.crt',
                'ssl_ciphers': 'EECDH+ECDSA+AESGCM:EECDH+aRSA+AESGCM:EECDH+ECDSA+SHA384:EECDH+ECDSA+SHA256:EECDH+aRSA+SHA384:EECDH+aRSA+SHA256:!EECDH+aRSA+RC4:EECDH:EDH+aRSA:!RC4:!aNULL:!eNULL:!LOW:!3DES:!MD5:!EXP:!PSK:!SRP:!DSS:@STRENGTH'
            },
            'filesets': None
        }

        file = {
            'manager': {
                'log_file': '/var/log/dnstable-manager/dnstable-manager.log',
                'log_level': 'ERROR',
                'clean_tempfiles': {
                    'on_startup': True,
                    'on_exit': True
                }
            },
            'downloader': {
                'max_downloads': 42,
                'retry_timeout': 60
            },
            'filesets': {
                'dns': {
                    'uri': 'https://export.dnsdb.info/dnsdb-export/mtbl/dns.fileset',
                    'destination': '/srv/dnstable/mtbl',
                    'base': 'dns',
                    'extension': 'mtbl',
                    'frequency': 900
                },
                'dnssec': {
                    'uri': 'https://export.dnsdb.info/dnsdb-export/mtbl/dnssec.fileset',
                    'destination': '/srv/dnstable/mtbl',
                    'base': 'dnssec',
                    'extension': 'mtbl',
                    'frequency': 600
                }
            }
        }

        stream = {
            'downloader': {'max_downloads': 32},
            'manager': {'clean_tempfiles': {'on_startup': False}}
        }

        expected = {
            'manager': {
                'log_file': '/var/log/dnstable-manager/dnstable-manager.log',
                'log_level': 'ERROR',
                'syslog': False,
                'syslog_facility': 'USER',
                'clean_tempfiles': {
                    'on_startup': False,
                    'on_exit': True
                }
            },
            'downloader': {
                'max_downloads': 32,
                'download_timeout': 60,
                'retry_timeout': 60,
                'ssl_ca_file': '/etc/ssl/certs/ca-certificates.crt',
                'ssl_ciphers': 'EECDH+ECDSA+AESGCM:EECDH+aRSA+AESGCM:EECDH+ECDSA+SHA384:EECDH+ECDSA+SHA256:EECDH+aRSA+SHA384:EECDH+aRSA+SHA256:!EECDH+aRSA+RC4:EECDH:EDH+aRSA:!RC4:!aNULL:!eNULL:!LOW:!3DES:!MD5:!EXP:!PSK:!SRP:!DSS:@STRENGTH'
            },
            'filesets': {
                'dns': {
                    'uri': 'https://export.dnsdb.info/dnsdb-export/mtbl/dns.fileset',
                    'destination': '/srv/dnstable/mtbl',
                    'base': 'dns',
                    'extension': 'mtbl',
                    'frequency': 900
                },
                'dnssec': {
                    'uri': 'https://export.dnsdb.info/dnsdb-export/mtbl/dnssec.fileset',
                    'destination': '/srv/dnstable/mtbl',
                    'base': 'dnssec',
                    'extension': 'mtbl',
                    'frequency': 600
                }
            }
        }

        actual = du.merge_dicts([default, file, stream])
        self.assertEqual(actual, expected)

class TestTempFile(unittest.TestCase):
    def setup_tempfile(self):
        self.temp_file = du.TempFile(prefix='dm-test-tempfile-', dir='/tmp')

    def test_has_correct_prefix(self):
        self.setup_tempfile()
        self.assertTrue(os.path.basename(self.temp_file.name).startswith('dm-test-tempfile-'))

    def test_is_in_correct_dir(self):
        self.setup_tempfile()
        test_path = os.path.join('/tmp', os.path.basename(self.temp_file.name))
        self.assertTrue(os.path.isfile(test_path))

    def test_has_correct_contents(self):
        self.setup_tempfile()
        temp_file_name = self.temp_file.name
        expected = [
            'rrnames.20210323.1730.m.tgdb',
            'rrnames.20210323.1731.m.tgdb',
            'rrnames.20210323.1732.m.tgdb',
            'rrnames.20210323.1733.m.tgdb',
            'rrnames.20210323.1734.m.tgdb',
            'rrnames.20210323.1735.m.tgdb',
            'rrnames.20210323.1736.m.tgdb'
        ]
        self.temp_file.write(('\n'.join(expected) + '\n').encode('utf-8'))
        self.temp_file.close()
        self.temp_file.delete = False
        with open(temp_file_name) as f:
            actual = [l.rstrip() for l in f.readlines()]
        self.assertEqual(actual, expected)
        # Clean up by deleting it.
        os.unlink(temp_file_name)

    def test_deleted_on_del(self):
        self.setup_tempfile()
        temp_file_name = self.temp_file.name
        del self.temp_file
        self.assertFalse(os.path.isfile(temp_file_name))

    def test_deleted_when_func_finishes(self):

        def run(dir):
            temp_file = du.TempFile(prefix='dm-test-tempfile-', dir=dir)
            temp_file.write('Hello Exception\n'.encode())
            temp_file.flush()
            self.assertTrue(len(os.listdir(t_dir)) == 1)

        with tempfile.TemporaryDirectory(prefix='dm-test-tempdir-', dir='/tmp') as t_dir:
            t1 = threading.Thread(target=run, args=(t_dir,), daemon=True)
            t1.start()
            t1.join()
            # Check that the dir is empty
            self.assertTrue(len(os.listdir(t_dir)) == 0)

    def test_not_deleted_when_done(self):
        self.setup_tempfile()
        temp_file_name = self.temp_file.name
        self.temp_file.delete = False
        del self.temp_file
        self.assertTrue(os.path.isfile(temp_file_name))
        # Clean up by deleting it.
        os.unlink(temp_file_name)

    def test_as_context_happy_path(self):
        expected = 'You must construct additional pylons\n'
        destination = '/tmp/dm-test-final-file'
        with du.TempFile(prefix='dm-test-tempfile-context', dir='/tmp', destination=destination) as tf:
            temp_file_name = tf.name
            tf.write(expected.encode('utf-8'))
        # Check that the tempfile has been renamed
        self.assertFalse(os.path.isfile(temp_file_name))
        # Test that the destination exists and has the right contents
        with open(destination) as f:
            self.assertEqual(f.read(), expected)
        # Clean up by deleting it.
        os.unlink(destination)

    def test_as_context_bad_write(self):
        # __exit__ should return False which raises the TypeError
        with self.assertRaises(TypeError):
            with du.TempFile(prefix='dm-test-tempfile-context', dir='/tmp') as tf:
                temp_file_name = tf.name
                tf.write('Not Bytes')

        # The file should not be here
        self.assertFalse(os.path.isfile(temp_file_name))

    def test_as_context_bad_destination(self):
        destination = '/nope-no-dir-here'
        try:
            with du.TempFile(prefix='dm-test-tempfile-context', dir='/tmp', destination=destination) as tf:
                temp_file_name = tf.name
                self.assertTrue(os.path.isfile(temp_file_name))
                tf.write('Hello World'.encode('utf-8'))
        except OSError:
            pass
        # The file should not be here
        self.assertFalse(os.path.isfile(temp_file_name))


class TestGetTotalContentLength(unittest.TestCase):
    def setUp(self):
        self.actual_length = 1000
        self.tell = 100
        self.sent_content_length = 900
        self.content_range = 'bytes 100-899/' + str(self.actual_length)

    def test_no_content_range(self):
        self.content_range = None
        actual = du.get_total_content_length(self.sent_content_length, self.tell,
                                             self.content_range)
        self.assertEqual(actual, self.actual_length)

    def test_with_content_range(self):
        actual = du.get_total_content_length(self.sent_content_length, self.tell,
                                             self.content_range)
        self.assertEqual(actual, self.actual_length)

    def test_not_resumed(self):
        self.content_range = None
        self.tell = 0
        self.sent_content_length = self.actual_length
        actual = du.get_total_content_length(self.sent_content_length, self.tell,
                                             self.content_range)
        self.assertEqual(actual, self.actual_length)

    def test_bad_content_range(self):
        self.content_range = 'foobar'
        actual = du.get_total_content_length(self.sent_content_length, self.tell,
                                             'foobar')
        self.assertEqual(actual, self.actual_length)

    def test_bad_content_range_string(self):
        self.content_range = self.content_range + 'abcd'
        actual = du.get_total_content_length(self.sent_content_length, self.tell,
                                             self.content_range)
        self.assertEqual(actual, self.actual_length)


class TestTerminatableThread(unittest.TestCase):
    def test_terminate(self):
        def foo():
            while True:
                pass

        x = du.TerminatableThread(target=foo, daemon=True)
        x.start()
        time.sleep(0.1)
        self.assertTrue(x.is_alive())
        x.terminate()
        time.sleep(0.1)
        self.assertFalse(x.is_alive())
